//  Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.apache.hivemind.service.impl;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.hivemind.ApplicationRuntimeException;
import org.apache.hivemind.HiveMindMessages;
import org.apache.hivemind.InterceptorStack;
import org.apache.hivemind.ServiceInterceptorFactory;
import org.apache.hivemind.internal.Module;
import org.apache.hivemind.service.ClassFab;
import org.apache.hivemind.service.ClassFabUtils;
import org.apache.hivemind.service.ClassFactory;
import org.apache.hivemind.service.MethodSignature;

/**
 * Base class for creating new service interceptors.  Most implementations
 * merely have to implement 
 * {@link #addServiceMethodImplementation(ClassFab, String, Class, Class[], Class[])}.
 * This is only suitable for interceptors which do not use parameters.
 * 
 * <p>
 * Implementations of this service must be configured with:
 * <ul>
 * <li>The service point id assigned to the <code>serviceId</code> property
 * <li>The <code>hivemind.ClassFactory</code> service assigned to the <code>factory</code> property
 * </ul>
 *
 * @author Howard Lewis Ship
 */
public abstract class AbstractServiceInterceptorFactory implements ServiceInterceptorFactory
{
    private ClassFactory _factory;
    private String _serviceId;
    private Map _cachedClasses;

    {
        _cachedClasses = Collections.synchronizedMap(new HashMap());
    }

    /**
     * Creates the interceptor. Expects that the parameters list is empty.
     * The class that is created is cached; if an interceptor is requested
     * for the same extension point, then the previously constructed class
     * is reused (this can happen with the threaded service model, for example,
     * when a thread-local service implementation is created for different threads).
     */
    public void createInterceptor(
        InterceptorStack stack,
        Module contributingModule,
        List parameters)
    {
        Class serviceInterfaceClass = stack.getServiceInterface();
        Class interceptorClass = getInterceptorClass(stack, parameters);

        try
        {
            Object interceptor = instantiateInterceptor(stack, interceptorClass, parameters);

            stack.push(interceptor);
        }
        catch (Exception ex)
        {
            throw new ApplicationRuntimeException(
                ServiceMessages.errorInstantiatingInterceptor(
                    _serviceId,
                    stack,
                    interceptorClass,
                    ex),
                ex);
        }
    }

    private Class getInterceptorClass(InterceptorStack stack, List parameters)
    {
        String id = stack.getServiceExtensionPointId();
        Class result = (Class) _cachedClasses.get(id);

        if (result != null)
            return result;

        result = constructInterceptorClass(stack, parameters);

        _cachedClasses.put(id, result);

        return result;
    }

    private Class constructInterceptorClass(InterceptorStack stack, List parameters)
    {
        Class serviceInterfaceClass = stack.getServiceInterface();
        Module module = stack.getServiceModule();

        String name = ClassFabUtils.generateClassName("Interceptor");

        ClassFab classFab = _factory.newClass(name, getInterceptorSuperclass(), module);

        classFab.addInterface(serviceInterfaceClass);

        createInfrastructure(stack, classFab, parameters);

        addServiceMethods(stack, classFab, parameters);

        return classFab.createClass();
    }

    /**
     * Overridden in subclasses to identify the super-class
     * for the interceptor.  This implementation returns java.lang.Object.
     */

    protected Class getInterceptorSuperclass()
    {
        return Object.class;
    }

    /**
     * Invoked in subclasses to create any infrastructure. 
     * <p>This implementation adds a field, <code>_inner</code> whose
     * type is the <em>actual</em> type of the 
     * {@link InterceptorStack#peek() top object on the stack} (which, of
     * course, implements the service interface).  A constructor is created
     * to set the field. 
     * 
     * @param stack the interceptor stack (used to obtain the service interface and extension point id)
     * @param classFab the interceptor class being constructed
     * @param parameters parameters provided to the interceptor factory 
     */
    protected void createInfrastructure(InterceptorStack stack, ClassFab classFab, List parameters)
    {
        Class topClass = stack.peek().getClass();

        classFab.addField("_inner", topClass);
        classFab.addConstructor(new Class[] { topClass }, null, "_inner = $1;");
    }

    /**
     * Used to instantiate the interceptor.  This implementation passes the top object
     * on the interceptor stack to the interceptorClass' constructor. Subclasses that
     * defined a different constructor (in {@link #createInfrastructure(InterceptorStack, ClassFab)})
     * will need to override this method as well.
     * 
     * @param stack the interceptor stack on which the returned interceptor will be placed.
     * @param interceptorClass the generated class for the interceptor.
     * @param the parameters provided to the interceptor
     * 
     * @throws Exception if there is an error getting or invoking the constructor
     * @see #createInfrastructure(InterceptorStack, ClassFab)
     */
    protected Object instantiateInterceptor(
        InterceptorStack stack,
        Class interceptorClass,
        List parameters)
        throws Exception
    {
        // Should only be a single constructor in a custom fabricated class.
        Constructor c = interceptorClass.getConstructors()[0];

        return c.newInstance(new Object[] { stack.peek()});
    }

    /**
     * Invoked by {@link #addServiceMethods(InterceptorStack, ClassFab, List)}
     * for each method in the service interface to allow the factory to
     * construct the corresponding method in the interceptor. This implementation
     * does nothing, and must be overridden. More sophisticated
     * interceptors will override <code>addServiceMethods()</code> and may by
     * implemented in such a way that this method is not invoked. 
     * 
     * @param classFab the fabricator for the interceptor class
     * @param sig the method signature for the method
     */

    protected void addServiceMethodImplementation(
        ClassFab classFab,
        MethodSignature sig)
    {
        throw new ApplicationRuntimeException(
            HiveMindMessages.unimplementedMethod(this, "addServiceMethodImplementation"));
    }

    /**
     * Invoked from {@link #createInterceptor(InterceptorStack, Module, List) to add
     * the service methods to the interceptor class. This implementation ignores
     * the parameters, and invokes 
     * {@link #addServiceMethodImplementation(ClassFab, String, Class, Class[], Class[])}
     * for each method in the service interface. It also may invoke
     * {@link #addToStringMethod(InterceptorStack, ClassFab)}.
     * 
     * 
     */
    protected void addServiceMethods(InterceptorStack stack, ClassFab fab, List parameters)
    {
        boolean toString = false;
        Method[] methods = stack.getServiceInterface().getMethods();

        for (int i = 0; i < methods.length; i++)
        {
            Method m = methods[i];

            addServiceMethodImplementation(fab, new MethodSignature(m));

            toString |= ClassFabUtils.isToString(m);
        }

        if (!toString)
            addToStringMethod(stack, fab);
    }

    /**
     * Creates a toString() method that identify the interceptor service id,
     * the intercepted service id, and the service interface class name).
     */
    protected void addToStringMethod(InterceptorStack stack, ClassFab fab)
    {
        ClassFabUtils.addToStringMethod(
            fab,
            "<Interceptor: "
                + _serviceId
                + " for "
                + stack.getServiceExtensionPointId()
                + "("
                + stack.getServiceInterface().getName()
                + ")>");

    }

    public void setServiceId(String string)
    {
        _serviceId = string;
    }

    public void setFactory(ClassFactory factory)
    {
        _factory = factory;
    }

}
