//  Copyright 2004 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package hivemind.test.services.impl;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hivemind.ApplicationRuntimeException;
import org.apache.hivemind.InterceptorStack;
import org.apache.hivemind.ServiceInterceptorFactory;
import org.apache.hivemind.internal.Module;
import org.apache.hivemind.service.BodyBuilder;
import org.apache.hivemind.service.ClassFab;
import org.apache.hivemind.service.ClassFabUtils;
import org.apache.hivemind.service.ClassFactory;
import org.apache.hivemind.service.MethodSignature;
import org.apache.hivemind.service.impl.AbstractLoggingInterceptor;

/**
 * Much like {@link org.apache.hivemind.service.impl.LoggingInterceptorFactory}, 
 * but uses parameters to limit the methods that will be logged.
 *
 * @author Howard Lewis Ship
 */
public class FilterLoggingInterceptor implements ServiceInterceptorFactory
{
    private ClassFactory _classFactory;

    public void createInterceptor(InterceptorStack stack, Module invokingModule, List parameters)
    {
        Set names = new HashSet();

        Iterator i = parameters.iterator();
        while (i.hasNext())
        {
            MethodFilter filter = (MethodFilter) i.next();
            names.add(filter.getName());
        }

        Class serviceInterfaceClass = stack.getServiceInterface();

        String name = ClassFabUtils.generateClassName("Interceptor");
        ClassFab classFab =
            _classFactory.newClass(
                name,
                AbstractLoggingInterceptor.class,
                stack.getServiceModule());

        classFab.addInterface(serviceInterfaceClass);
        classFab.addField("_inner", serviceInterfaceClass);

        classFab.addConstructor(
            new Class[] { Log.class, serviceInterfaceClass },
            null,
            "{ super($1); _inner = $2; }");

        Method[] methods = serviceInterfaceClass.getMethods();

        BodyBuilder builder = new BodyBuilder();

        for (int j = 0; j < methods.length; j++)
        {
            Method m = methods[j];
            String methodName = m.getName();

            builder.clear();

            builder.begin();

            if (names.contains(methodName))
            {
                builder.add("_logEntry(");
                builder.addQuoted(methodName);
                builder.addln(", $args);");
            }

            builder.add("return ($r) ");
            builder.add("_inner.");
            builder.add(methodName);
            builder.addln("($$);");

            builder.end();

            classFab.addMethod(Modifier.PUBLIC, new MethodSignature(m), builder.toString());

        }

        Class interceptorClass = classFab.createClass();

        Log log = LogFactory.getLog(stack.getServiceExtensionPointId());

        Object interceptor = null;

        try
        {
            Constructor c =
                interceptorClass.getConstructor(new Class[] { Log.class, serviceInterfaceClass });

            interceptor = c.newInstance(new Object[] { log, stack.peek()});
        }
        catch (Exception ex)
        {
            throw new ApplicationRuntimeException(ex);
        }

        stack.push(interceptor);
    }

    public void setClassFactory(ClassFactory factory)
    {
        _classFactory = factory;
    }

}
